// #############################################################################
// #                     --- Flashlight Controller ---                         #
// #############################################################################
// # main.c - Main program                                                     #
// #############################################################################
// #              Version: 1.0 - Compiler: AVR-GCC 4.5.0 (Linux)               #
// #  (c) 2012 by Malte Pöggel - www.MALTEPOEGGEL.de - malte@maltepoeggel.de   #
// #############################################################################
// #  This program is free software; you can redistribute it and/or modify it  #
// #   under the terms of the GNU General Public License as published by the   #
// #        Free Software Foundation; either version 3 of the License,         #
// #                  or (at your option) any later version.                   #
// #                                                                           #
// #      This program is distributed in the hope that it will be useful,      #
// #      but WITHOUT ANY WARRANTY; without even the implied warranty of       #
// #           MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.            #
// #           See the GNU General Public License for more details.            #
// #                                                                           #
// #  You should have received a copy of the GNU General Public License along  #
// #      with this program; if not, see <http://www.gnu.org/licenses/>.       #
// #############################################################################

 #include <avr/io.h>
 #include <avr/pgmspace.h>
 #include <avr/eeprom.h>
 #include <util/delay.h>
 #include "main.h"
 #include "prng.h"

 // Current active mode
 volatile uint8_t mode;
 uint8_t eeMode EEMEM = 0;
 
 // Active modeset
 volatile uint8_t modes;
 uint8_t eeModes EEMEM = 5;
 
 // Start with last state (0), always start with brightest level (1)
 volatile uint8_t start;
 uint8_t eeStart EEMEM = 0;

 // Modeset counter
 volatile uint8_t mcount;
 uint8_t eeMcount EEMEM = 0;


 // Main program
 int main( void )
  {
   // Variables
   uint8_t temp_a = 0;
   uint8_t temp_b = 0;
   uint8_t temp_c = 0;
   uint8_t bctr   = 0;
   uint16_t bvolt = 0;
   uint8_t cctr   = 40;
  
   // Setup ports
   DDRB |= (1<<PB0);  // OC0A -> PWM
   DDRB &= ~(1<<PB3); // ADC3 <- BAT

   // Setup timer
   TCCR0A |= (1<<COM0A1) | (1<<WGM00);
   TCCR0B |= (1<<CS00);
   OCR0A = 0;

   // Setup ADC
   ADCSRA |= (1<<ADEN) | (1<<ADPS2) | (1<<ADPS1); // Clock prescaler 64
   ADMUX |= (1<<REFS1) | (1<<MUX1) | (1<<MUX0); // vref=1.1V, input=ADC3

   // ADC dummy readout
   ADCSRA |= (1<<ADSC);
   while(ADCSRA & (1<<ADSC));
   bvolt = ADCL | (ADCH<<8);

   // Debounce
   _delay_ms(5);

   // Read and validate eeprom data
   modes = eeprom_read_byte(&eeModes);
   if(modes>=NO_MODES)
    {
     modes = 0;
     eeprom_write_byte(&eeModes, modes);
    }
   mode = eeprom_read_byte(&eeMode);
   if(mode>=pgm_read_byte(&mode_array[modes][0]))
    {
     mode = 0;
     eeprom_write_byte(&eeMode, mode);
    }
   start = eeprom_read_byte(&eeStart);
   if(start>1)
    {
     start=0;
     eeprom_write_byte(&eeStart, start);
    }
   mcount = eeprom_read_byte(&eeMcount);
   if(mcount>25)
    {
     mcount=0;
     eeprom_write_byte(&eeMcount, mcount);
    }

   // Check reset type
   if(MCUSR&(1<<PORF))
   {
    // Power on reset.
    MCUSR &= ~(1<<PORF); // Clear flag
     if(start!=0&&mode!=0) // Always start with first mode?
      {
       mode = 0;
       eeprom_write_byte(&eeMode, mode); // Store in eep
      }
     if(mcount!=0)
      {
       mcount = 0;
       eeprom_write_byte(&eeMcount, mcount);
      }
    } else {
     // This is no power on reset.
     if(++mode>=pgm_read_byte(&mode_array[modes][0])) // Increment mode, check overflow
      {
       mode=0;
      }
     eeprom_write_byte(&eeMode, mode); // Store in eep
     if(++mcount<=25)
      {
       eeprom_write_byte(&eeMcount, mcount);
      }
    }

   // Main loop
   for(;;)
    {

     if(cctr>0)
      {
       if(cctr==1)
        {
         // Check clicks for mode switch
         if(mcount==20) // Next modeset
          {
           mode = 0;
           if(++modes>=NO_MODES) modes = 0;
           eeprom_write_byte(&eeModes, modes);
           eeprom_write_byte(&eeMode, mode);
           mcount = 0;
           eeprom_write_byte(&eeMcount, mcount);
           OCR0A = 0x00;
           _delay_ms(500);
           for(temp_a=0; temp_a<(1+modes); temp_a++)
            {
             OCR0A = 0xFF;
             _delay_ms(250);
             OCR0A = 0x11;
             _delay_ms(500);
            }
           OCR0A = 0x00;   
           while(1) {}
          }
         if(mcount==25) // Toggle last state / always brightest level
          {
           if(++start>1) start = 0;
           eeprom_write_byte(&eeStart, start);
           mcount = 0;
           eeprom_write_byte(&eeMcount, mcount);      
           OCR0A = 0x00;
           _delay_ms(500);
           for(temp_a=0; temp_a<(NO_MODES+1+start); temp_a++)
            {
             OCR0A = 0xFF;
             _delay_ms(250);
             OCR0A = 0x11;
             _delay_ms(500);
            }
           OCR0A = 0x00;
           while(1) {}
          }
         if(mcount!=0) // Reset if count incorrect
          {
           mcount = 0;
           eeprom_write_byte(&eeMcount, mcount);
          }
        }
       cctr--;
      }

     // Active Mode
     switch(pgm_read_byte(&mode_array[modes][(mode+1)]))
      {
       case MODE_BRIGHT_7:
        OCR0A = 0xFF;
        break;
       case MODE_BRIGHT_6:
        OCR0A = 0xCC;
        break;
       case MODE_BRIGHT_5:
        OCR0A = 0x99;
        break;
       case MODE_BRIGHT_4:
        OCR0A = 0x66;
        break;
       case MODE_BRIGHT_3:
        OCR0A = 0x33;
        break;
       case MODE_BRIGHT_2:
        OCR0A = 0x22;
        break;
       case MODE_BRIGHT_1:
        OCR0A = 0x11;
        break;
       case MODE_BLINK:
        // Blink
        if(temp_a<=10) OCR0A = 255; else OCR0A = 0;
        if(++temp_a>=20) temp_a=0;
        break;
       case MODE_STROBE:
        // Strobe
        if(temp_a<=1) OCR0A = 255; else OCR0A = 0;
        if(++temp_a>=4) temp_a=0;
        break;
       case MODE_FLASH:
        // Short flash 2sec
        if(temp_a==0) OCR0A = 255; else OCR0A = 0;
        if(++temp_a>=40) temp_a=0;
        break;
       case MODE_SOS:
        // SOS
        if(pgm_read_byte(&sos_pattern[temp_a/8])&(1<<(temp_a%8))) OCR0A = 255; else OCR0A = 0;
        if(++temp_a>=236) temp_a=0;
        break;
       case MODE_ALPINE:
        // Alpine distress
        if(++temp_a>=200)
         {
          temp_b++;
          temp_a=0;
         }
        if(temp_b>=12)
         {
          temp_b=0;
         }
        if(temp_b<6&&temp_a<16) OCR0A = 255; else OCR0A = 0;
        break;
       case MODE_FADE:
        // Fade up/down
        if(temp_b==0)
         {
          temp_a+=5;
          if(temp_a==255) temp_b=1;
         } else {
          temp_a-=5;
          if(temp_a==0) temp_b=0;
         }
        OCR0A=temp_a;
        break;
       case MODE_SPOOKY:
        // Spooky
        if(++temp_a>=5)
         {
          temp_a=0;
          temp_b=prng_get();
         }
        if(temp_b<temp_c)
         {
          if((temp_c-temp_b)>=10) temp_c-=10; else temp_c--;
         }
        if(temp_b>temp_c)
         {
          if((temp_b-temp_c)>=10) temp_c+=10; else temp_c++;
         }
        OCR0A=temp_c;
        break;
      }
      
     // Undervoltage cutoff
     ADCSRA |= (1<<ADSC);
     while(ADCSRA & (1<<ADSC));
     bvolt = ADCL | (ADCH<<8);
     if(bvolt<BATT_MIN)
      {
       if(++bctr>=10)
        {
         OCR0A = 0;
         ADCSRA &= ~(1<<ADEN);
         while(1) {};
        }
      } else {
       bctr=0;
      }
     
     // Delay for main loop 
     _delay_ms(50);
    }

   return 0;
  }
